/*
	\author Matt English
	\date 1/11/07
	\brief
		This file contains the function declarations for the NodeMemManager class.

		All content copyright  2006 DigiPen(USA) Corporation
*/
#include "RayCaster.h"

extern Scene* myScene;

#ifndef PI
#define PI 3.14159265
#endif
/*!
	The interface function called by the main loop
	
	\param i
		x position of the pixel we are raytracing from

	\param j
		y position of the pixel we are raytracing from

	\param width
		width of the viewport

	\param height
		height of the viewport

	\return
		Vector containing the color of the pixed

	\author
		Matt English
*/
Vector RayCast(float i, float j, int width, int height, bool bSoftShadows)
{
	Vector rayPos, rayDir, finalColor;
	GetWorldRay(i, j, width, height, rayPos, rayDir);
	finalColor = CheckCollision(rayPos, rayDir, 0, NULL, bSoftShadows);

	//cap the color values to black and white
	if(finalColor.x > 1.0f)
		finalColor.x = 1.0f;
	else if(finalColor.x < 0)
		finalColor.x = 0;
	if(finalColor.y > 1.0f)
		finalColor.y = 1.0f;
	else if(finalColor.y < 0)
		finalColor.y = 0;
	if(finalColor.z > 1.0f)
		finalColor.z = 1.0f;
	else if(finalColor.z < 0)
		finalColor.z = 0;

	return(finalColor);
}
/*!
	The interface function called by the main loop
	
	\param rayPos
		the position of the ray we are checking

	\param rayDir
		the direction of the ray we are checking

	\param recDepth
		the currect depth of recursion we are at

	\param shapeIn
		pointer to the shape the ray originates in if there is one, NULL otherwise

	\param bSoftShadows
		softShadow flag

	\return
		Vector containing the color of the pixed

	\author
		Matt English
*/
Vector CheckCollision(Vector &rayPos, Vector &rayDir, unsigned int recDepth, BaseShape *shapeIn, bool bSoftShadows)
{
	int length, i;
	float dist=(float)LARGE_FLOAT;
	float tempDist;
	Vector norm, intersection, finalNorm;
	Vector finalColor(0,0,0);
	BaseShape* shape = NULL;

	//check against all spheres
	length = (int)myScene->sphereList.size();
	Sphere *curSphere;
	for(i=0; i < length; ++i)
	{
		curSphere = &(myScene->sphereList[i]);
		tempDist = curSphere->Intersect(rayPos, rayDir, norm);
		if(tempDist != -1.0f && tempDist < dist)
		{
			dist = tempDist;
			finalNorm = norm;
			shape = curSphere;
		}
	}

	//check against all boxes
	length = (int)myScene->boxList.size();
	Box *curBox;
	for(i=0; i < length; ++i)
	{
		curBox = &(myScene->boxList[i]);
		tempDist = curBox->Intersect(rayPos, rayDir, norm);
		if(tempDist != -1.0f && tempDist < dist)
		{
			dist = tempDist;
			finalNorm = norm;
			shape = curBox;
		}
	}

	//now check against all polys
	length = (int)myScene->polyList.size();
	Poly *curPoly;
	for(i=0; i < length; ++i)
	{
		curPoly = &(myScene->polyList[i]);
		tempDist = curPoly->Intersect(rayPos, rayDir, norm);
		if(tempDist != -1.0f && tempDist < dist)
		{
			dist = tempDist;
			finalNorm = norm;
			shape = curPoly;
		}
	}

	//now check against all ellipsoids
	length = (int)myScene->ellipsoidList.size();
	Ellipsoid *curEllipsoid;
	for(i=0; i < length; ++i)
	{
		curEllipsoid = &(myScene->ellipsoidList[i]);
		tempDist = curEllipsoid->Intersect(rayPos, rayDir, norm);
		if(tempDist != -1.0f && tempDist < dist)
		{
			dist = tempDist;
			finalNorm = norm;
			shape = curEllipsoid;
		}
	}

	//now check against all lenses
	length = (int)myScene->lensList.size();
	ConvexLens *curLens;
	for(i=0; i < length; ++i)
	{
		curLens = &(myScene->lensList[i]);
		tempDist = curLens->Intersect(rayPos, rayDir, norm);
		if(tempDist != -1.0f && tempDist < dist)
		{
			dist = tempDist;
			finalNorm = norm;
			shape = curLens;
		}
	}

	//now do local light calculation if we have hit an object
	if(shape != NULL)
	{
		intersection = rayPos + (rayDir * dist);
		finalColor = CalculateColor(intersection, rayDir, finalNorm.Normalize(), shape, recDepth, shapeIn, dist, bSoftShadows);
	}
	return(finalColor);
}
/*!
	Creates a ray in world space from the pixel coordinates
	
	\param x
		x position of the pixel we are raytracing from

	\param y
		y position of the pixel we are raytracing from

	\param width
		width of the viewport

	\param height
		height of the viewport

	\param outPos
		position of the vector in world space

	\param outDir
		direction of the vector in world space

	\author
		Matt English
*/
void GetWorldRay(float x, float y, int width, int height, Vector& outPos, Vector& outDir)
{
	//using hansons notes
	//mapped x and y
	float mx, my;
	//map x any y to between -1 and 1
	mx = (float)((float)((float)x/(float)width) * 2.0f) - 1.0f;
	my = (float)((float)((float)y/(float)height) * 2.0f) - 1.0f;
	Vector target = myScene->viewPlaneCenter +
		(myScene->viewPlaneU * (float)mx) + (myScene->viewPlaneV * (float)my);
	outPos = myScene->camera;
	outDir = target - outPos;

}
/*! Calculates the color of the light at a specified point on an object

    \param intersection
		The position in world where we are calculating the diffuse color

    \param norm
		The normal of the object at point intersection

	\param shape
		A BaseShape pointer to the object
    
	\param recDepth
		The recursive depth

	\param shapeIn
		pointer to the shape the ray originates in

    \return
		COLORREF of the color at that point taking the diffuse light into account
*/
Vector CalculateColor(Vector intersection, Vector rayDir, Vector norm, BaseShape* shape, unsigned int recDepth,
					  BaseShape *shapeIn, float dist, bool bSoftShadows)
{
	Vector finalColor, lightDir, reflection, transmission;
	float numerator, temp, reflectFraction, shadowRatio;
	
	rayDir = rayDir.Normalize();

	finalColor.x = finalColor.y = finalColor.z = 0;

	float erPar, erPerp, eiPar, eiPerp;
	float cosAngle, refractRatio, mPermRatio, radicand;
	//calculate the refaction ratio and the magnetic permiability ratio based on whether we are inside a shape
	if(shapeIn == NULL)
	{
		refractRatio = 1.0f/shape->refractionIndex;
		mPermRatio = 1.0f/shape->relMagnPermeability;
	}
	else
	{
		refractRatio = shapeIn->refractionIndex;
		mPermRatio = shapeIn->relMagnPermeability;
	}
	
	//calculate transmission color

	temp = rayDir * norm;
	radicand = 1.0f - (pow(refractRatio, 2) * (1.0f - pow(fabs(temp), 2)));
	if(radicand > EPSILON)
	{
		radicand = sqrt(radicand);
		cosAngle = fabs(temp);
		float srt = mPermRatio * radicand;
		//solve the fresnel equations
		erPerp = (refractRatio * cosAngle) - srt;
		eiPerp = (refractRatio * cosAngle) + srt;

		srt = refractRatio * radicand;
		erPar = (mPermRatio * cosAngle) - srt;
		eiPar = (mPermRatio * cosAngle) + srt;

		reflectFraction = 0.5f * (pow(erPerp/eiPerp, 2) + pow(erPar/eiPar, 2));
	}
	else
		reflectFraction = 1.0f;

	shadowRatio = 0.0f;
	//we must go through each light
	if(shapeIn == NULL)
	{
		for(std::vector<Light>::iterator iter=myScene->lightList.begin(); iter != myScene->lightList.end(); ++iter)
		{
			lightDir = iter->point - intersection;
			shadowRatio += CheckShadow(intersection, lightDir, norm, bSoftShadows);
			//skip if completely in shadow
			if(shadowRatio <= 0.0f)
				continue;
			lightDir = lightDir.Normalize();
			//distance squared between the intersection point and the light source
			numerator = lightDir * norm;
			if(numerator < 0)
				numerator = 0;
			//for each color
			finalColor.x += ((shape->diffusecolor.x * numerator) * iter->color.x);
			finalColor.y += ((shape->diffusecolor.y * numerator) * iter->color.y);
			finalColor.z += ((shape->diffusecolor.z * numerator) * iter->color.z);
			
			//specular component
			//first get the reflection vector of the light about the normal
			float tempDot = lightDir * norm;
			if(tempDot < 0)
				tempDot = 0;
			reflection = ((norm * 2.0f * (tempDot)) - lightDir);
			tempDot = reflection * -rayDir;
			//if the dot product is positive, we have specular here
			if(tempDot >= 0)
				finalColor += iter->color * (pow(tempDot, shape->specularReflectionExp) * shape->specularReflectionCoef * reflectFraction);
		}
		shadowRatio /= myScene->lightList.size();
		finalColor = finalColor * shadowRatio;
		//now add the ambient
		finalColor.x += myScene->ambient.x * shape->diffusecolor.x;
		finalColor.y += myScene->ambient.y * shape->diffusecolor.y;
		finalColor.z += myScene->ambient.z * shape->diffusecolor.z;
	}
	if(recDepth < maxReflectionDepth)
	{
		float capT = 1.0f - reflectFraction;
		Vector newColor;
		//dont calculate negligible values
		if(reflectFraction > R_EPSILON)
		{
			//calculate the relection component
			reflection = (rayDir - (norm * 2.0f * (rayDir * norm)));
			//if we are in a shape
			if(shapeIn != NULL)
				newColor = CheckCollision(intersection - norm * (float)SHADOW_EPSILON, reflection, recDepth + 1, shapeIn, bSoftShadows);
			//not in a shape, push the intersection point out
			else
				newColor = CheckCollision(intersection + norm * (float)SHADOW_EPSILON, reflection, recDepth + 1, shapeIn, bSoftShadows);
			finalColor += (newColor * shape->specularReflectionCoef * reflectFraction);
		}

		if(capT > R_EPSILON)
		{
			temp = -rayDir * norm;
			//get the new ray for transmission
			if(temp >= 0)
				radicand = -radicand;

			transmission = (norm * (radicand + (temp * refractRatio))) - (-rayDir * refractRatio);
			if(shapeIn != NULL)
				newColor = CheckCollision(intersection + norm * (float)SHADOW_EPSILON, transmission, recDepth + 1, NULL, bSoftShadows);
			else
				newColor = CheckCollision(intersection - norm * (float)SHADOW_EPSILON, transmission, recDepth + 1, shape, bSoftShadows);
			finalColor += (newColor * shape->specularReflectionCoef * capT);
		}
	}
	//apply the attenuation factors
	if(shapeIn == NULL && myScene->air.x != -1)
	{
		finalColor.x *= pow(myScene->air.x, dist);
		finalColor.y *= pow(myScene->air.y, dist);
		finalColor.z *= pow(myScene->air.z, dist);
	}
	else if(shapeIn != NULL)
	{
		finalColor.x *= pow(shapeIn->transAttenFactors.x, dist);
		finalColor.y *= pow(shapeIn->transAttenFactors.y, dist);
		finalColor.z *= pow(shapeIn->transAttenFactors.z, dist);
	}
	return(finalColor);
}
/*! Determines whether the current intersection point is shadowed

    \param intersection
		The position in world where we are calculating the color

    \param dir
		The non-normalized direction vector from the intersection point to the light
    
	\param norm
		normal to the object at the point of intersection

    \return
		true if there is an object in between the intersection point and the light point,
	false otherwise
*/
float CheckShadow(Vector intersection, Vector dir, Vector norm, bool bSoftShadows)
{
	/*add the epsilon to the intersection point so we do not collide with
	the object again right near our original point if floating point error
	would cause that*/
	intersection = intersection + norm * (float)SHADOW_EPSILON;
	if(bSoftShadows == false)
	{
		if(SimpleCheckCollision(intersection, dir) == true)
			return(0.0f);
		else
			return(1.0f);
	}
	else
	{
		int col = 0;
		Vector newDir = dir;
		for(int i=0; i < SOFT_RAYS; ++i)
		{
			//peturb the ray to create cool softshadows
			newDir.x += ((rand()/RAND_MAX) - 0.5f) * 0.01f;
			newDir.y += ((rand()/RAND_MAX) - 0.5f) * 0.01f;
			newDir.z += ((rand()/RAND_MAX) - 0.5f) * 0.01f;
			if(SimpleCheckCollision(intersection, newDir) == true)
				++col;
		}
		return(1.0f - ((float)col / (float)SOFT_RAYS));
	}
}
/*!
	Simplified check collision function for checking wheter a point is in shadow.  This seperate
	function is used over a dual use function for speed and code simplicity.
	
	\param rayPos
		the position of the ray we are checking

	\param rayDir
		the direction of the ray we are checking

	\return bool
		true if there is an object in the way, false otherwise		

	\author
		Matt English
*/
bool SimpleCheckCollision(Vector rayPos, Vector rayDir)
{
	int length, i;
	float tempDist;
	Vector norm;

	length = (int)myScene->sphereList.size();
	Sphere *curSphere;
	for(i=0; i < length; ++i)
	{
		curSphere = &(myScene->sphereList[i]);
		tempDist = curSphere->Intersect(rayPos, rayDir, norm);
		if(tempDist != -1.0f && tempDist < 1.0f)
		{
			return(true);
		}
	}

	//check against all boxes
	length = (int)myScene->boxList.size();
	Box *curBox;
	for(i=0; i < length; ++i)
	{
		curBox = &(myScene->boxList[i]);
		tempDist = curBox->Intersect(rayPos, rayDir, norm);
		if(tempDist != -1.0f && tempDist < 1.0f)
		{
			return(true);
		}
	}

	//now check against all polys
	length = (int)myScene->polyList.size();
	Poly *curPoly;
	for(i=0; i < length; ++i)
	{
		curPoly = &(myScene->polyList[i]);
		tempDist = curPoly->Intersect(rayPos, rayDir, norm);
		if(tempDist != -1.0f && tempDist < 1.0f)
		{
			return(true);
		}
	}

	//now check against all ellipsoids
	length = (int)myScene->ellipsoidList.size();
	Ellipsoid *curEllipsoid;
	for(i=0; i < length; ++i)
	{
		curEllipsoid = &(myScene->ellipsoidList[i]);
		tempDist = curEllipsoid->Intersect(rayPos, rayDir, norm);
		if(tempDist != -1.0f && tempDist < 1.0f)
		{
			return(true);
		}
	}

	//now check against all lenses
	length = (int)myScene->lensList.size();
	ConvexLens *curLens;
	for(i=0; i < length; ++i)
	{
		curLens = &(myScene->lensList[i]);
		tempDist = curLens->Intersect(rayPos, rayDir, norm);
		if(tempDist != -1.0f && tempDist < 1.0f)
		{
			return(true);
		}
	}
	return(false);
}